/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.crypto;

import java.nio.charset.StandardCharsets;
import java.security.Provider;
import java.security.SecureRandom;
import java.security.Security;

import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.storage.SecretValue;

/**
 * Encrypts and decrypts secret data.
 */
public class SecretTransformer {
  private static final Logger logger = LoggerFactory.getLogger(SecretTransformer.class);

  private final Cryptor cryptor;

  public static SecretTransformer getSecretTransformer(String masterEncryptionKey) {
    return new SecretTransformer(
        new Cryptor(getSecretKey(masterEncryptionKey), getProvider(), new SecureRandom())
    );
  }

  public SecretTransformer(Cryptor cryptor) {
    this.cryptor = cryptor;
  }

  public Secret transform(SecretValue secretValue) {
    if (secretValue == null) {
      return null;
    }
    return new Secret(
        secretValue.getPath(),
        secretValue.getKey(),
        secretValue.getVersion(),
        new String(cryptor.decrypt(secretValue.getEncrypted()), StandardCharsets.UTF_8)
    );
  }

  public SecretValue transform(Secret secretEntity) {
    if (secretEntity == null) {
      return null;
    }
    byte[] bytes = secretEntity.getSecret().getBytes(StandardCharsets.UTF_8);
    return new SecretValue(
        secretEntity.getPath(),
        secretEntity.getKey(),
        secretEntity.getVersion(),
        cryptor.encryptionKeyDerivedFrom(secretEntity.getPath() + "/" + secretEntity.getKey())
            .encrypt(bytes),
        cryptor.computeHmac(bytes),
        null,
        System.currentTimeMillis()
    );
  }

  private static Provider getProvider() {
    Provider provider = new BouncyCastleProvider();
    if (Security.getProvider(provider.getName()) == null) {
      logger.debug("Registering new crypto provider {}", provider.getName());
      Security.addProvider(provider);
    }
    return provider;
  }

  private static SecretKey getSecretKey(String masterKey) {
    return new SecretKeySpec(masterKey.getBytes(StandardCharsets.UTF_8), "AES");
  }
}
