/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.storage;

import java.io.IOException;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.metrics.JmxReporter;
import org.apache.kafka.common.metrics.MetricConfig;
import org.apache.kafka.common.metrics.Metrics;
import org.apache.kafka.common.metrics.MetricsReporter;
import org.apache.kafka.common.metrics.Sensor;
import org.apache.kafka.common.metrics.stats.Value;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.apache.kafka.common.config.ConfigException;
import org.apache.kafka.common.MetricName;
import org.apache.kafka.common.utils.Time;
import io.confluent.kafka.secretregistry.client.rest.RestService;
import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.client.rest.entities.requests.RegisterSecretRequest;
import io.confluent.kafka.secretregistry.client.rest.exceptions.RestClientException;
import io.confluent.kafka.secretregistry.client.rest.utils.UrlList;
import io.confluent.kafka.secretregistry.crypto.SecretTransformer;
import io.confluent.kafka.secretregistry.exceptions.IdGenerationException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryInitializationException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryRequestForwardingException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryStoreException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryTimeoutException;
import io.confluent.kafka.secretregistry.exceptions.UnknownMasterException;
import io.confluent.kafka.secretregistry.masterelector.kafka.KafkaGroupMasterElector;
import io.confluent.kafka.secretregistry.rest.SecretRegistryConfig;
import io.confluent.kafka.secretregistry.rest.SslFactory;
import io.confluent.kafka.secretregistry.rest.VersionId;
import io.confluent.kafka.secretregistry.rest.exceptions.Errors;
import io.confluent.kafka.secretregistry.storage.exceptions.StoreException;
import io.confluent.kafka.secretregistry.storage.exceptions.StoreInitializationException;
import io.confluent.kafka.secretregistry.storage.exceptions.StoreTimeoutException;
import io.confluent.kafka.secretregistry.storage.serialization.SecretRegistrySerializer;
import io.confluent.kafka.secretregistry.storage.serialization.Serializer;
import io.confluent.rest.Application;
import io.confluent.rest.RestConfig;
import io.confluent.rest.exceptions.RestException;

public class KafkaSecretRegistry implements SecretRegistry, MasterAwareSecretRegistry {

  public static final int MIN_VERSION = 1;
  public static final int MAX_VERSION = Integer.MAX_VALUE;
  private static final Logger log = LoggerFactory.getLogger(KafkaSecretRegistry.class);
  private static final Pattern LISTENER_PATTERN =
      Pattern.compile("^(.*)://\\[?([0-9a-zA-Z\\-%._:]*)\\]?:(-?[0-9]+)");

  private final SecretRegistryConfig config;
  private final LookupCache<SecretRegistryKey, SecretRegistryValue> lookupCache;
  // visible for testing
  final KafkaStore<SecretRegistryKey, SecretRegistryValue> kafkaStore;
  private final Serializer<SecretRegistryKey, SecretRegistryValue> serializer;
  private final int kafkaStoreTimeoutMs;
  private final int initTimeout;
  private final boolean isEligibleForMasterElector;
  private SecretTransformer secretTransformer;

  private SecretRegistryIdentity myIdentity;
  private final Object masterLock = new Object();
  private SecretRegistryIdentity masterIdentity;
  private RestService masterRestService;
  private SslFactory sslFactory;
  private MasterElector masterElector = null;
  private Metrics metrics;
  private Sensor masterNodeSensor;

  public KafkaSecretRegistry(SecretRegistryConfig config) throws SecretRegistryException {
    if (config == null) {
      throw new SecretRegistryException("Schema registry configuration is null");
    }
    this.config = config;
    this.isEligibleForMasterElector = config.getBoolean(SecretRegistryConfig.MASTER_ELIGIBILITY);
    this.kafkaStoreTimeoutMs = config.getInt(SecretRegistryConfig.KAFKASTORE_TIMEOUT_CONFIG);
    this.initTimeout = config.getInt(SecretRegistryConfig.KAFKASTORE_INIT_TIMEOUT_CONFIG);
    this.serializer = new SecretRegistrySerializer();
    this.lookupCache = lookupCache();
    this.kafkaStore = kafkaStore(config);
    this.secretTransformer = SecretTransformer.getSecretTransformer(config.masterEncryptionKey());
  }

  protected KafkaStore<SecretRegistryKey, SecretRegistryValue> kafkaStore(
      SecretRegistryConfig config
  ) throws SecretRegistryException {
    return new KafkaStore<SecretRegistryKey, SecretRegistryValue>(
        config,
        groupId(config),
        new KafkaStoreMessageHandler(this, lookupCache),
        this.serializer, lookupCache, new NoopKey()
    );
  }

  // visible for testing
  protected static String groupId(SecretRegistryConfig config) {
    return config.getString(SecretRegistryConfig.KAFKASTORE_GROUP_ID_CONFIG).isEmpty()
        ? String.format("secret-registry-%s",
        config.getString(SecretRegistryConfig.HOST_NAME_CONFIG)
    )
        : config.getString(SecretRegistryConfig.KAFKASTORE_GROUP_ID_CONFIG);
  }

  protected LookupCache<SecretRegistryKey, SecretRegistryValue> lookupCache() {
    return new InMemoryCache<SecretRegistryKey, SecretRegistryValue>();
  }

  @Override
  public SecretRegistryConfig config() {
    return config;
  }

  @Override
  public void initStore() throws SecretRegistryException {
    try {
      kafkaStore.init();
      maybeRotateKeys();
    } catch (StoreInitializationException | StoreException e) {
      throw new SecretRegistryInitializationException(
          "Error initializing kafka store while initializing secret registry", e);
    }
  }

  private void maybeRotateKeys() throws StoreException {
    String oldKey = config.masterEncryptionOldKey();
    if (oldKey != null) {
      log.info("Rotating master key");
      // Ignore any errors during decoding since old secret may not
      // have been removed during restart.
      SecretTransformer oldTransformer = SecretTransformer.getSecretTransformer(oldKey);
      Iterator<SecretRegistryKey> allKeys = kafkaStore.getAllKeys();
      try {
        while (allKeys.hasNext()) {
          SecretRegistryKey secretKey = allKeys.next();
          SecretRegistryValue secretValue = kafkaStore.get(secretKey);
          if (secretValue instanceof SecretValue) {
            Secret secret = oldTransformer.transform((SecretValue) secretValue);
            SecretValue newSecretValue = secretTransformer.transform(secret);
            kafkaStore.put(secretKey, newSecretValue);
          }
        }
      } catch (Exception e) {
        log.info("Could not rotate, ignoring old master key");
      }
      log.info("Done rotating master key");
    }
  }

  /**
   * A Secret Registry instance's identity is in part the port it listens on. Currently the port can
   * either be configured via the deprecated `port` configuration, or via the `listeners`
   * configuration.
   *
   * <p>This method uses `Application.parseListeners()` from `rest-utils` to get a list of
   * listeners, and returns the port of the first listener to be used for the instance's identity.
   *
   * <p></p>In theory, any port from any listener would be sufficient. Choosing the first, instead
   * of say the last, is arbitrary.
   */
  public static UriInfo getUriInfoForIdentity(
      String host,
      Integer port,
      List<String> configuredListeners,
      String requestedScheme
  ) throws SecretRegistryException {
    try {
      if (host == null) {
        InetAddress localhost = InetAddress.getLocalHost();
        host = localhost.getHostAddress();
      }
      if (port == null) {
        port = SecretRegistryConfig.SECRETREGISTRY_PORT_DEFAULT;
      }
      if (configuredListeners == null) {
        return new UriInfo(requestedScheme, host, port);
      }
      List<UriInfo> listeners = parseListeners(configuredListeners,
          host,
          port,
          Arrays.asList(SecretRegistryConfig.HTTP, SecretRegistryConfig.HTTPS),
          SecretRegistryConfig.HTTP
      );
      for (UriInfo listener : listeners) {
        if (requestedScheme.equalsIgnoreCase(listener.scheme())) {
          return listener;
        }
      }
    } catch (Exception e) {
      throw new SecretRegistryException("Could not obtain URI info", e);
    }
    throw new SecretRegistryException("No listener configured with requested scheme "
        + requestedScheme);
  }

  public static List<UriInfo> parseListeners(
      List<String> listenersConfig,
      String host,
      Integer deprecatedPort,
      List<String> supportedSchemes,
      String defaultScheme
  ) {
    // handle deprecated case, using PORT_CONFIG.
    // TODO: remove this when `PORT_CONFIG` is deprecated, because LISTENER_CONFIG
    // will have a default value which includes the default port.
    if (listenersConfig.isEmpty() || listenersConfig.get(0).isEmpty()) {
      log.warn(
          "DEPRECATION warning: `listeners` configuration is not configured. "
              + "Falling back to the deprecated `port` configuration."
      );
      listenersConfig = new ArrayList<String>(1);
      listenersConfig.add(defaultScheme + "://" + host + ":" + deprecatedPort);
    }

    List<UriInfo> listeners = new ArrayList<UriInfo>(listenersConfig.size());
    for (String listenerStr : listenersConfig) {
      Matcher listenerMatcher = LISTENER_PATTERN.matcher(listenerStr);

      if (!listenerMatcher.matches()) {
        throw new ConfigException("Listener doesn't have the right format "
            + "(protocol://hostname:port).");
      }

      String protocol = listenerMatcher.group(1).toLowerCase(Locale.ENGLISH);

      String hostname = listenerMatcher.group(2);
      if (hostname == null || hostname.isEmpty() || hostname.equals("0.0.0.0")) {
        hostname = host;
      }
      int port = Integer.parseInt(listenerMatcher.group(3));

      if (!supportedSchemes.contains(protocol)) {
        log.warn(
            "Found a listener with an unsupported scheme (supported: {}). Ignoring listener '{}'",
            supportedSchemes,
            listenerStr
        );
      } else {
        listeners.add(new UriInfo(protocol, hostname, port));
      }
    }

    if (listeners.isEmpty()) {
      throw new ConfigException("No listeners are configured. Must have at least one listener.");
    }

    return listeners;
  }

  @Override
  public void initRest(SslFactory sslFactory, UriInfo uriInfo) throws SecretRegistryException {
    try {
      log.info("Initializing REST with URI {}", uriInfo);
      this.sslFactory = sslFactory;
      this.myIdentity = new SecretRegistryIdentity(
              uriInfo.host(), uriInfo.port(), isEligibleForMasterElector, uriInfo.scheme()
      );

      MetricConfig metricConfig =
              new MetricConfig().samples(config.getInt(ProducerConfig.METRICS_NUM_SAMPLES_CONFIG))
                      .timeWindow(
                              config.getLong(ProducerConfig.METRICS_SAMPLE_WINDOW_MS_CONFIG),
                              TimeUnit.MILLISECONDS
                      );
      List<MetricsReporter> reporters =
              config.getConfiguredInstances(
                      ProducerConfig.METRIC_REPORTER_CLASSES_CONFIG,
                      MetricsReporter.class
              );
      String jmxPrefix = "kafka.secret.registry";
      reporters.add(new JmxReporter(jmxPrefix));
      this.metrics = new Metrics(metricConfig, reporters, Time.SYSTEM);
      this.masterNodeSensor = metrics.sensor("master-slave-role");

      Map<String, String> configuredTags = Application.parseListToMap(
              config.getList(RestConfig.METRICS_TAGS_CONFIG)
      );
      MetricName
              m = new MetricName("master-slave-role", "master-slave-role",
              "1.0 indicates the node is the active master in the cluster and is the"
                      + " node where all register secret requests are "
                      + "served.", configuredTags
      );
      this.masterNodeSensor.add(m, new Value());

      log.info("Joining secret registry with Kafka-based coordination");
      masterElector = new KafkaGroupMasterElector(config, myIdentity, this);
      masterElector.init();
    } catch (SecretRegistryStoreException e) {
      throw new SecretRegistryInitializationException(
              "Error electing master while initializing secret registry", e);
    } catch (SecretRegistryTimeoutException e) {
      throw new SecretRegistryInitializationException(e);
    }
  }

  public boolean isMaster() {
    synchronized (masterLock) {
      if (masterIdentity != null && masterIdentity.equals(myIdentity)) {
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * 'Inform' this SecretRegistry instance which SecretRegistry is the current master.
   * If this instance is set as the new master, ensure it is up-to-date with data in
   * the kafka store.
   *
   * @param newMaster Identity of the current master. null means no master is alive.
   */
  @Override
  public void setMaster(SecretRegistryIdentity newMaster)
      throws SecretRegistryTimeoutException, SecretRegistryStoreException, IdGenerationException {
    log.debug("Setting the master to " + newMaster);

    // Only secret registry instances eligible for master can be set to master
    if (newMaster != null && !newMaster.getMasterEligibility()) {
      throw new IllegalStateException(
          "Tried to set an ineligible node to master: " + newMaster);
    }

    synchronized (masterLock) {
      SecretRegistryIdentity previousMaster = masterIdentity;
      masterIdentity = newMaster;

      if (masterIdentity == null) {
        masterRestService = null;
      } else {
        masterRestService = new RestService(masterIdentity.getUrl());
        if (sslFactory != null && sslFactory.sslContext() != null) {
          masterRestService.setSslSocketFactory(sslFactory.sslContext().getSocketFactory());
        }
      }

      if (masterIdentity != null && !masterIdentity.equals(previousMaster) && isMaster()) {
        // The new master may not know the exact last offset in the Kafka log. So, mark the
        // last offset invalid here
        kafkaStore.markLastWrittenOffsetInvalid();
        //ensure the new master catches up with the offsets before it gets nextid and assigns
        // master
        try {
          kafkaStore.waitUntilKafkaReaderReachesLastOffset(initTimeout);
        } catch (StoreException e) {
          throw new SecretRegistryStoreException("Exception getting latest offset ", e);
        }
      }
      masterNodeSensor.record(isMaster() ? 1.0 : 0.0);
    }
  }

  /**
   * Return json data encoding basic information about this SecretRegistry instance, such as
   * host, port, etc.
   */
  public SecretRegistryIdentity myIdentity() {
    return myIdentity;
  }

  /**
   * Return the identity of the SecretRegistry that this instance thinks is current master.
   * Any request that requires writing new data gets forwarded to the master.
   */
  public SecretRegistryIdentity masterIdentity() {
    synchronized (masterLock) {
      return masterIdentity;
    }
  }

  @Override
  public int register(Secret secret)
      throws SecretRegistryException {
    log.info("Registering secret '{}:{}'", secret.getPath(), secret.getKey());
    try {
      // Ensure cache is up-to-date before any potential writes
      kafkaStore.waitUntilKafkaReaderReachesLastOffset(kafkaStoreTimeoutMs);

      // determine the latest version of the secret for the path/key
      Iterator<Secret> allVersions = getAllVersions(secret.getPath(), secret.getKey()).iterator();

      int newVersion = MIN_VERSION;
      while (allVersions.hasNext()) {
        newVersion = allVersions.next().getVersion() + 1;
      }
      secret.setVersion(newVersion);

      SecretValue secretValue = secretTransformer.transform(secret);
      kafkaStore.put(new SecretKey(secret.getPath(), secret.getKey(), secret.getVersion()),
          secretValue);
      return newVersion;
    } catch (StoreTimeoutException te) {
      throw new SecretRegistryTimeoutException("Write to the Kafka store timed out while", te);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException("Error while registering the secret in the"
          + " backend Kafka store", e);
    }
  }

  public int registerOrForward(
      Secret secret,
      Map<String, String> headerProperties
  )
      throws SecretRegistryException {
    synchronized (masterLock) {
      if (isMaster()) {
        return register(secret);
      } else {
        // forward registering request to the master
        if (masterIdentity != null) {
          return forwardRegisterRequestToMaster(secret, headerProperties);
        } else {
          throw new UnknownMasterException("Register secret request failed since master is "
              + "unknown");
        }
      }
    }
  }

  @Override
  public void deleteVersion(String path, String key, int version)
      throws SecretRegistryException {
    log.info("Deleting secret '{}:{}' version {}", path, key, version);
    try {
      // Ensure cache is up-to-date before any potential writes
      kafkaStore.waitUntilKafkaReaderReachesLastOffset(kafkaStoreTimeoutMs);
      kafkaStore.delete(new SecretKey(path, key, version));
    } catch (StoreTimeoutException te) {
      throw new SecretRegistryTimeoutException("Write to the Kafka store timed out while", te);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException("Error while deleting the secret in the"
          + " backend Kafka store", e);
    }
  }

  public void deleteVersionOrForward(
      Map<String, String> headerProperties, String path, String key, int version
  )
      throws SecretRegistryException {

    synchronized (masterLock) {
      if (isMaster()) {
        deleteVersion(path, key, version);
      } else {
        // forward registering request to the master
        if (masterIdentity != null) {
          forwardDeleteVersionRequestToMaster(headerProperties, path, key, version);
        } else {
          throw new UnknownMasterException("Register secret request failed since master is "
              + "unknown");
        }
      }
    }
  }

  @Override
  public List<Integer> deleteKey(String path, String key) throws SecretRegistryException {
    log.info("Deleting key '{}:{}'", path, key);
    // Ensure cache is up-to-date before any potential writes
    try {
      List<Integer> versions = new ArrayList<>();
      kafkaStore.waitUntilKafkaReaderReachesLastOffset(kafkaStoreTimeoutMs);
      List<Secret> secrets = getAllVersions(path, key);
      for (Secret secret : secrets) {
        deleteVersion(secret.getPath(), secret.getKey(), secret.getVersion());
        versions.add(secret.getVersion());
      }
      return versions;
    } catch (StoreTimeoutException te) {
      throw new SecretRegistryTimeoutException("Write to the Kafka store timed out while", te);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException("Error while deleting the key in the"
          + " backend Kafka store", e);
    }
  }

  public List<Integer> deleteKeyOrForward(
      Map<String, String> requestProperties,
      String path, String key
  ) throws SecretRegistryException {
    synchronized (masterLock) {
      if (isMaster()) {
        return deleteKey(path, key);
      } else {
        // forward registering request to the master
        if (masterIdentity != null) {
          return forwardDeleteKeyRequestToMaster(requestProperties, path, key);
        } else {
          throw new UnknownMasterException("Register secret request failed since master is "
              + "unknown");
        }
      }
    }
  }

  @Override
  public Set<String> deletePath(String path) throws SecretRegistryException {
    log.info("Deleting path '{}'", path);
    // Ensure cache is up-to-date before any potential writes
    try {
      Set<String> keys = new HashSet<>();
      kafkaStore.waitUntilKafkaReaderReachesLastOffset(kafkaStoreTimeoutMs);
      SecretKey key1 = new SecretKey(path, SecretKey.MIN_KEY, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, SecretKey.MAX_KEY, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      while (allVersions.hasNext()) {
        Secret secret = secretTransformer.transform((SecretValue) allVersions.next());
        deleteVersion(secret.getPath(), secret.getKey(), secret.getVersion());
        keys.add(secret.getKey());
      }
      return keys;
    } catch (StoreTimeoutException te) {
      throw new SecretRegistryTimeoutException("Write to the Kafka store timed out while", te);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException("Error while deleting the path in the"
          + " backend Kafka store", e);
    }
  }

  public Set<String> deletePathOrForward(
      Map<String, String> requestProperties,
      String path
  ) throws SecretRegistryException {
    synchronized (masterLock) {
      if (isMaster()) {
        return deletePath(path);
      } else {
        // forward registering request to the master
        if (masterIdentity != null) {
          return forwardDeletePathRequestToMaster(requestProperties, path);
        } else {
          throw new UnknownMasterException("Register secret request failed since master is "
              + "unknown");
        }
      }
    }
  }


  private int forwardRegisterRequestToMaster(
      Secret secret,
      Map<String, String> headerProperties
  )
      throws SecretRegistryRequestForwardingException {
    final UrlList baseUrl = masterRestService.getBaseUrls();

    RegisterSecretRequest registerSecretRequest = new RegisterSecretRequest();
    registerSecretRequest.setSecret(secret.getSecret());
    log.debug(String.format("Forwarding registering secret request %s to %s",
        registerSecretRequest, baseUrl
    ));
    try {
      int id = masterRestService.registerSecret(
          headerProperties, registerSecretRequest, secret.getPath(), secret.getKey());
      return id;
    } catch (IOException e) {
      throw new SecretRegistryRequestForwardingException(
          String.format("Unexpected error while forwarding the registering secret request %s to %s",
              registerSecretRequest, baseUrl
          ),
          e
      );
    } catch (RestClientException e) {
      throw new RestException(e.getMessage(), e.getStatus(), e.getErrorCode(), e);
    }
  }

  private int forwardDeleteVersionRequestToMaster(
      Map<String, String> headerProperties,
      String path, String key,
      Integer version
  ) throws SecretRegistryRequestForwardingException {
    UrlList baseUrl = masterRestService.getBaseUrls();

    log.debug(String.format("Forwarding deleteSecretVersion secret version request %s-%s-%s to %s",
        path, key, version, baseUrl
    ));
    try {
      return masterRestService.deleteVersion(headerProperties, path, key, String.valueOf(version));
    } catch (IOException e) {
      throw new SecretRegistryRequestForwardingException(
          String.format(
              "Unexpected error while forwarding deleteSecretVersion secret version "
                  + "request %s-%s-%s to %s", path, key, version, baseUrl), e);
    } catch (RestClientException e) {
      throw new RestException(e.getMessage(), e.getStatus(), e.getErrorCode(), e);
    }
  }

  private List<Integer> forwardDeleteKeyRequestToMaster(
      Map<String, String> requestProperties,
      String path, String key
  ) throws SecretRegistryRequestForwardingException {
    UrlList baseUrl = masterRestService.getBaseUrls();

    log.debug(String.format("Forwarding delete key request for %s-%s to %s",
        path, key, baseUrl
    ));
    try {
      return masterRestService.deleteKey(requestProperties, path, key);
    } catch (IOException e) {
      throw new SecretRegistryRequestForwardingException(
          String.format(
              "Unexpected error while forwarding delete key "
                  + "request %s-%s to %s", path, key, baseUrl), e);
    } catch (RestClientException e) {
      throw new RestException(e.getMessage(), e.getStatus(), e.getErrorCode(), e);
    }
  }

  private Set<String> forwardDeletePathRequestToMaster(
      Map<String, String> requestProperties,
      String path
  ) throws SecretRegistryRequestForwardingException {
    UrlList baseUrl = masterRestService.getBaseUrls();

    log.debug(String.format("Forwarding delete path request for %s to %s",
        path, baseUrl
    ));
    try {
      return masterRestService.deletePath(requestProperties, path);
    } catch (IOException e) {
      throw new SecretRegistryRequestForwardingException(
          String.format(
              "Unexpected error while forwarding delete path "
                  + "request %s to %s", path, baseUrl), e);
    } catch (RestClientException e) {
      throw new RestException(e.getMessage(), e.getStatus(), e.getErrorCode(), e);
    }
  }

  public Secret validateAndGetSecret(String path, String key, VersionId versionId)
      throws SecretRegistryException {
    Secret secret = this.get(path, key, versionId.getVersionId());
    if (secret == null) {
      if (!this.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      } else if (!this.listKeys(path).contains(key)) {
        throw Errors.keyNotFoundException();
      } else {
        throw Errors.versionNotFoundException();
      }
    }
    return secret;
  }

  @Override
  public Secret get(String path, String key, int version)
      throws SecretRegistryException {
    log.info("Getting secret '{}:{}' version {}", path, key, version);
    VersionId versionId = new VersionId(version);
    if (versionId.isLatest()) {
      return getLatestVersion(path, key);
    } else {
      SecretKey secretKey = new SecretKey(path, key, version);
      try {
        SecretValue secretValue = (SecretValue) kafkaStore.get(secretKey);
        Secret secret = secretTransformer.transform(secretValue);
        return secret;
      } catch (StoreException e) {
        throw new SecretRegistryStoreException(
            "Error while retrieving secret from the backend Kafka"
                + " store", e);
      }
    }
  }

  @Override
  public Set<String> listPaths() throws SecretRegistryException {
    log.info("Listing all paths");
    try {
      Iterator<SecretRegistryKey> allKeys = kafkaStore.getAllKeys();
      return extractUniquePaths(allKeys);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  private Set<String> extractUniquePaths(Iterator<SecretRegistryKey> allKeys)
      throws StoreException {
    Set<String> paths = new HashSet<String>();
    while (allKeys.hasNext()) {
      SecretRegistryKey k = allKeys.next();
      if (k instanceof SecretKey) {
        SecretKey key = (SecretKey) k;
        SecretValue value = (SecretValue) kafkaStore.get(key);
        paths.add(key.getPath());
      }
    }
    return paths;
  }

  @Override
  public Set<String> listKeys(String path) throws SecretRegistryException {
    log.info("Listing keys for '{}'", path);
    try {
      SecretKey key1 = new SecretKey(path, SecretKey.MIN_KEY, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, SecretKey.MAX_KEY, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      return extractUniqueKeys(allVersions);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  private Set<String> extractUniqueKeys(Iterator<SecretRegistryValue> allVersions) {
    Set<String> keys = new HashSet<String>();
    while (allVersions.hasNext()) {
      SecretRegistryValue v = allVersions.next();
      if (v instanceof SecretValue) {
        SecretValue value = (SecretValue) v;
        keys.add(value.getKey());
      }
    }
    return keys;
  }

  @Override
  public List<Integer> listVersions(String path, String key) throws SecretRegistryException {
    log.info("Listing versions for '{}:{}'", path, key);
    try {
      SecretKey key1 = new SecretKey(path, key, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, key, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      return extractUniqueVersions(allVersions);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  private List<Integer> extractUniqueVersions(Iterator<SecretRegistryValue> allVersions) {
    List<Integer> versions = new ArrayList<>();
    while (allVersions.hasNext()) {
      SecretRegistryValue v = allVersions.next();
      if (v instanceof SecretValue) {
        SecretValue value = (SecretValue) v;
        versions.add(value.getVersion());
      }
    }
    return versions;
  }

  @Override
  public List<Secret> getAllLatest(String path)
      throws SecretRegistryException {
    log.info("Getting latest for '{}'", path);
    try {
      SecretKey key1 = new SecretKey(path, SecretKey.MIN_KEY, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, SecretKey.MAX_KEY, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      List<Secret> secrets = new ArrayList<>(latestKeys(path, allVersions).values());
      Collections.sort(secrets);
      return secrets;
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  @Override
  public List<Secret> getAllVersions(String path, String key)
      throws SecretRegistryException {
    log.info("Getting versions for '{}:{}'", path, key);
    try {
      SecretKey key1 = new SecretKey(path, key, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, key, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      return sortSecrets(allVersions);
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  @Override
  public Secret getLatestVersion(String path, String key) throws SecretRegistryException {
    try {
      SecretKey key1 = new SecretKey(path, key, MIN_VERSION);
      SecretKey key2 = new SecretKey(path, key, MAX_VERSION);
      Iterator<SecretRegistryValue> allVersions = kafkaStore.getAll(key1, key2);
      List<Secret> sortedVersions = sortSecrets(allVersions);
      Secret latestSecret = null;
      if (sortedVersions.size() > 0) {
        latestSecret = sortedVersions.get(sortedVersions.size() - 1);
      }
      return latestSecret;
    } catch (StoreException e) {
      throw new SecretRegistryStoreException(
          "Error from the backend Kafka store", e);
    }
  }

  @Override
  public void close() {
    log.info("Shutting down secret registry");
    kafkaStore.close();
    if (masterElector != null) {
      masterElector.close();
    }
  }

  KafkaStore<SecretRegistryKey, SecretRegistryValue> getKafkaStore() {
    return this.kafkaStore;
  }

  private List<Secret> sortSecrets(Iterator<SecretRegistryValue> secrets) {
    List<Secret> secretVector = new ArrayList<>();
    while (secrets.hasNext()) {
      SecretValue secretValue = (SecretValue) secrets.next();
      secretVector.add(secretTransformer.transform(secretValue));
    }
    Collections.sort(secretVector);
    return secretVector;
  }

  private Map<String, Secret> latestKeys(String path, Iterator<SecretRegistryValue> secrets) {
    Map<String, Secret> result = new HashMap<>();
    Map<String, Integer> maxVersions = new HashMap<>();
    while (secrets.hasNext()) {
      SecretValue secretValue = (SecretValue) secrets.next();
      Secret secret = secretTransformer.transform(secretValue);
      String key = secret.getKey();
      int version = secret.getVersion();
      int value = maxVersions.merge(key, version, Math::max);
      if (version == value) {
        result.put(key, secret);
      }
    }
    return result;
  }
}
